#!/usr/bin/env python
"""
_ExpressInjectorComponent_

Component for generating Express jobs
"""

__version__ = "$Revision: 1.24 $"
__revision__ = "$Id: ExpressInjectorComponent.py,v 1.24 2010/07/20 22:53:32 hufnagel Exp $"

import os
import shutil
import logging
import sys

from JobQueue import JobQueueDB

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0.WorkflowFactory.ExpressWorkflow import ExpressWorkflow
from T0.JobFactory.StreamerJob import StreamerJob

from T0.State.Database.Writer import InsertRun

from T0.ExpressInjector.ExpressInjectorAPI import defineJobs
from T0.ExpressInjector.ExpressInjectorAPI import registerJobs

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec

from ProdAgentCore.Configuration import loadProdAgentConfiguration
from T0.GenericTier0 import Tier0DB


class ExpressInjectorComponent(GenericTier0Component):
    """
    _ExpressInjectorComponent_

    Component for generating Express jobspecs.
    """
    def __init__(self, **args):
        GenericTier0Component.__init__(self, **args) 
        
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")
        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")
        self.jobCacheDir = self.lookupParameter("JobCacheDir", self.args["ComponentDir"])

        self.workflowSpecQueue = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2

        self.maxInputEvents = int(self.lookupParameter("MaxInputEvents", "100"))

        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        self.popconlogDBConn = None

        logging.info("ExpressInjector Component Started")


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component receiving a message
        """
        # always have a debug on/off switch
        if message == "ExpressInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "ExpressInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "ExpressInjector:Poll":
            self.poll()
        elif message == "ExpressInjector:CleanUp":
            self.cleanupRun(payload)
        else:
            logging.error("Unknown message: %s, payload: %s" % (message, payload))

        return

    def createWorkflow(self, runNumber, stream_id):
        """
        create the workflow template and publish message for a given run

        """
        # no need to check return value, already did this earlier
        runConfig = self.getRunConfig(runNumber)

        stream = runConfig.getStreamName(stream_id)

        logging.debug("Create new workflow for run %d and stream %s" % (runNumber, stream))

        expressConfig = runConfig.getExpressConfiguration(stream_id)
        if expressConfig == None:
            logging.error("No express config for run %d and stream %s" % (runNumber, stream))
            return None

        scenario = expressConfig.Scenario

        if expressConfig.ProcessingConfigURL == None:
            configFile = None
        else:
            configFile = runConfig.retrieveConfigFromURL(expressConfig.ProcessingConfigURL)

        # filter out ALCARECO from data tiers
        enableAlcaReco = False
        if "ALCARECO" in expressConfig.DataTiers:
            expressConfig.DataTiers.remove("ALCARECO")
            enableAlcaReco = True

        # filter out DQM from data tiers
        enableDQM = False
        if "DQM" in expressConfig.DataTiers:
            expressConfig.DataTiers.remove("DQM")
            enableDQM = True

        outModuleInfoList = []
        if expressConfig.SplitInProcessing:
            outModuleInfoList = runConfig.getOutputModuleInfoList("Express", stream_id,
                                                                  expressConfig.DataTiers,
                                                                  splitIntoDatasets = True)
        else:
            outModuleInfoList = runConfig.getOutputModuleInfoList("Express", stream_id,
                                                                  expressConfig.DataTiers)

        if outModuleInfoList == None:
            logging.error("No output module list for run %s." % runNumber)
            return None

        # now add the single output module for combined ALCARECO
        if enableAlcaReco:
            outModuleInfo = runConfig.getOutputModuleInfo("Express","ALCARECO",
                                                          streamID = stream_id)
            if outModuleInfo == None:
                logging.error("No output module for run %s and data tier ALCARECO" % runNumber)
                return None

            outModuleInfoList.append(outModuleInfo)

        # now add the single output module for combined DQM
        if enableDQM:
            outModuleInfo = runConfig.getOutputModuleInfo("Express","DQM",
                                                          streamID = stream_id)
            if outModuleInfo == None:
                logging.error("No output module for run %s and data tier DQM" % runNumber)
                return None

            outModuleInfoList.append(outModuleInfo)

        workflowDir = os.path.join(self.jobCacheDir, "workflows",
                                   "Run%s" % runNumber)
        workflowSpecFile = os.path.join(workflowDir,
                                        "Express-Run%s-Stream%s-workflow.xml" % (runNumber, stream))

        pid = os.fork()
        if pid != 0:
            os.wait()

            if not os.path.exists(workflowSpecFile):
                logging.debug("No workflow for run, stream: %s, %s" % (runNumber, stream))
                return None

            workflowSpec = WorkflowSpec()
            workflowSpec.load(workflowSpecFile)

            self.ms.publish("NewWorkflow", workflowSpecFile)
            self.ms.commit()

            return workflowSpec

        workflow = ExpressWorkflow(runNumber,
                                   expressConfig.CMSSWVersion,
                                   self.cmsPath,
                                   self.scramArch,
                                   *outModuleInfoList)

        workflowSpec = workflow.makeWorkflowSpec("Express-Run%s-Stream%s" % (runNumber, stream),
                                                 scenario = scenario,
                                                 configFile = configFile,
                                                 alcaProducers = expressConfig.AlcaProducers,
                                                 globalTag = expressConfig.GlobalTag,
                                                 enableLazyDownload = self.useLazyDownload,
                                                 productionMode = self.productionMode)

        if workflowSpec == None:
            sys.exit(0)
        
        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)
        
        workflowSpec.save(workflowSpecFile)
            
        logging.debug("Workflow file saved as %s" % workflowSpecFile)
        sys.exit(0)

    def createJobs(self):
        """
        _createJobs_

        main routine called in polling cycle

        """
        # define new jobs we need to create
        jobEntityList = defineJobs(self.t0astDBConn,
                                   self.maxInputEvents)

        # sort by run and stream_id
        jobEntityDict = {}
        for jobEntity in jobEntityList:
            run = jobEntity["runNumber"]
            stream_id = jobEntity["streamID"]
            if jobEntityDict.has_key(run):
                if jobEntityDict[run].has_key(stream_id):
                    jobEntityDict[run][stream_id].append(jobEntity)
                else:
                    jobEntityDict[run][stream_id] = [ jobEntity ]
            else:
                jobEntityDict[run] = {}
                jobEntityDict[run][stream_id] = [ jobEntity ]

        # now process by run and stream_id
        # only try to create new workflow once per run and stream_id
        # keep track of which jobs were created
        createdJobs = []
        createdJobSpecs = []
        for run in sorted(jobEntityDict.keys()):

            #
            # check that this run has a RunConfig record
            #
            runConfig = self.getRunConfig(run)

            if runConfig == None:
                logging.error("Could not retrieve run config for run %s" % run)
                continue

            jobSpecBaseDir = os.path.join(self.jobCacheDir,
                                          "jobspecs",
                                          "Run%d" % jobEntity["runNumber"])

            for stream_id in sorted(jobEntityDict[run].keys()):

                workflowSpecsByRun = self.workflowSpecQueue.get(run, None)
                if workflowSpecsByRun == None:
                    workflowSpecsByRun = {}
                    self.workflowSpecQueue[run] = workflowSpecsByRun

                workflowSpec = workflowSpecsByRun.get(stream_id, None)

                if workflowSpec == None:
                    workflowSpec = self.createWorkflow(run, stream_id)
                    if workflowSpec == None:
                        logging.error("Can't create workflow for run %d and stream %s"
                                      % (run, runConfig.getStreamName(stream_id)))
                        continue
                    else:
                        self.workflowSpecQueue[run][stream_id] = workflowSpec
                    
                # reuse job factory for all jobs in a run and stream
                jobFactory = StreamerJob(workflowSpec)
                for jobEntity in jobEntityDict[run][stream_id]:
                    jobSpec = jobFactory("Express", jobEntity, jobSpecBaseDir)
                    createdJobs.append(jobEntity)
                    createdJobSpecs.append(jobSpec)

        if len(createdJobs) > 0:

            try:
                self.paDBConn.beginTransaction()

                for jobSpec in createdJobSpecs:
                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters['JobSpecFile'],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                registerJobs(self.t0astDBConn, createdJobs)

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(createdJobs))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        return


    def poll(self):
        """
        _poll_

        Single cycle invocation of job creation

        """
        logging.info("poll(): Running...")

        # first check for new express released runs
        try:
            InsertRun.releaseRunsForExpress(self.t0astDBConn, self.popconlogDBConn)
        except StandardError, ex:
            logging.error("ERROR: Failed to release runs for express : %s\n" % str(ex))

        # then create new jobs
        self.createJobs()
        
        self.ms.publishUnique("ExpressInjector:Poll", "", self.pollInterval)
        self.ms.commit()
        return

    def cleanupRun(self, runNumber):
        """
        _cleanupRun_

        Remove all workflows and jobspecs for a particular run from disk.
        Also, remove any workflows for the run from memory.
        """
        try:
            runNumber = int(runNumber)
        except ValueError:
            logging.error("Run number is not valid: %s" % runNumber)
            return

        logging.debug("Cleaning up run %s" % runNumber)
        
        # remove workflow from queue
        if runNumber in self.workflowSpecQueue.keys():
            del self.workflowSpecQueue[runNumber]
            
        workflowDir = self.jobCacheDir + "/workflows/Run%s" % runNumber
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % runNumber
        
        shutil.rmtree(workflowDir,ignore_errors=True)
        shutil.rmtree(jobSpecDir,ignore_errors=True)
        return

    def startComponent(self):
        """
        _startComponent_

        Start up the component and define the messages that it subscribes to.
        """
        self.ms.subscribeTo("ExpressInjector:StartDebug")
        self.ms.subscribeTo("ExpressInjector:EndDebug")
        self.ms.subscribeTo("ExpressInjector:Poll")

        self.ms.subscribeTo("CleanUp")
        
        self.connectT0AST()
        self.connectPADB()

        paConfig = loadProdAgentConfiguration()
        popconlogDBConfig = paConfig.getConfig("PopConLogDB")
        if popconlogDBConfig == None:
            self.popconlogDBConn = None
        else:
            self.popconlogDBConn = Tier0DB.Tier0DB(popconlogDBConfig,
                                                   manageGlobal = False)
            self.popconlogDBConn.connect()

        self.ms.publishUnique("ExpressInjector:Poll", "", self.pollInterval)
        self.ms.commit()
        
        # wait for messages
        self.messageLoop()
