#!/usr/bin/env python
"""
_ExpressMerge_

Splitting algorithm for express merges.
"""

__revision__ = "$Id: ExpressMerge.py,v 1.3 2010/07/21 13:23:28 hufnagel Exp $"
__version__  = "$Revision: 1.3 $"

import time
import threading

from WMCore.WMBS.File import File

from WMCore.JobSplitting.JobFactory import JobFactory
from WMCore.Services.UUID import makeUUID
from WMCore.DAOFactory import DAOFactory


class ExpressMerge(JobFactory):
    """
    Split jobs by set of files

    """
    def algorithm(self, groupInstance = None, jobInstance = None, *args,
                  **kwargs):
        """
        _algorithm_

        A file based splitting algorithm

        """
        # extract some global scheduling parameters
        self.maxAlcaLumiAge = kwargs['maxAlcaLumiAge']
        self.maxDqmLumiAge = kwargs['maxDqmLumiAge']
        self.maxLumiAge = kwargs['maxLumiAge']
        self.maxInputFiles = kwargs['maxInputFiles']
        self.maxInputSize = kwargs['maxInputSize']
        self.currentTime = time.time()

        # flag to know whether a group has been created
        self.createdGroup = False

        myThread = threading.currentThread()
        daoFactory = DAOFactory(package = "T0.State.Database",
                                logger = myThread.logger,
                                dbinterface = myThread.dbi)

        getFilesDAO = daoFactory(classname = "ListFilesForExpressMerge")
        fileset = getFilesDAO.execute(self.subscription['id'])

        # sort files by run, stream, datatier, primds and procds
        mergeFilesDict = {}
        for f in fileset:

            # convenience reference
            run = f['run']
            stream_id = f['stream_id']
            datatier = f['datatier']
            primds_id = f['primds_id']
            procds_id = f['procds_id']

            # make sure the dictionary tree is initialized
            # we sort by run -> stream_id -> datatier -> primds -> procds
            if not mergeFilesDict.has_key(run):
                mergeFilesDict[run] = {}
            if not mergeFilesDict[run].has_key(stream_id):
                mergeFilesDict[run][stream_id] = {}
            if not mergeFilesDict[run][stream_id].has_key(datatier):
                mergeFilesDict[run][stream_id][datatier] = {}
            if not mergeFilesDict[run][stream_id][datatier].has_key(primds_id):
                mergeFilesDict[run][stream_id][datatier][primds_id] = {}
            if not mergeFilesDict[run][stream_id][datatier][primds_id].has_key(procds_id):
                mergeFilesDict[run][stream_id][datatier][primds_id][procds_id] = []

            mergeFilesDict[run][stream_id][datatier][primds_id][procds_id].append(f)

        # define merge jobs by run, stream, data tier, primary dataset and processed dataset
        for run in sorted(mergeFilesDict.keys()):
            for stream_id in mergeFilesDict[run].keys():
                for datatier in mergeFilesDict[run][stream_id].keys():
                    for primds_id in mergeFilesDict[run][stream_id][datatier].keys():
                        for mergeFiles in mergeFilesDict[run][stream_id][datatier][primds_id].values():
                            self.defineMergeJobs(mergeFiles, datatier)

        return


    def defineMergeJobs(self, mergeFiles, datatier):
        """
        _defineMergeJobs_

        Given a list of files that can be merged together, determine
        whether to issue merge jobs and how to assign files to them
        and in which order

        """
        myThread = threading.currentThread()
        myThread.logger.debug("defineMergeJobs(): Running...")

        # sort by lumi section
        mergeFilesByLumi = {}
        for mergeFile in mergeFiles:
            lumi = mergeFile['lumi']
            if not mergeFilesByLumi.has_key(lumi):
                mergeFilesByLumi[lumi] = []
            mergeFilesByLumi[lumi].append(mergeFile)

        #
        # actual scheduling algorith
        #
        # criteria:
        #     1. a lumi section should not sit around too long, waiting to get merged
        #     2. if possible, we want to merge in order of lumi sections (no holes)
        #     3. avoid too many small files (merge as many lumi sections as possible)
        #     4. don't merge too many lumi sections as the jobs run too long
        #     5. last, don't produce too big files
        #
        # some of these are mutually exclusive
        #
        # given up on creating single scheduling algorithm for everything
        #
        # ALCARECO:
        #     merge whatever we have if oldest lumi section older than self.maxAlcaLumiAge
        #     if self.maxAlcaLumiAge is 0, merge lumi by lumi
        #
        # DQM:
        #     merge whatever we have if oldest lumi section older than self.maxDqmLumiAge
        #     if self.maxDqmLumiAge is 0, merge lumi by lumi
        #
        # everything else:
        #     merge whatever we have if oldest lumi section older than self.maxLumiAge
        #     if self.maxLumiAge is 0, merge lumi by lumi
        #

        if datatier == "ALCARECO":
            timeout = self.maxAlcaLumiAge
        elif datatier == "DQM":
            timeout = self.maxDqmLumiAge
        else:
            timeout = self.maxLumiAge

        lastLumi = 0
        jobSizeTotal = 0
        jobFileList = []
        for lumi in sorted(mergeFilesByLumi.keys()):

            lumiFileList = mergeFilesByLumi[lumi]
            lumiAge = self.currentTime - lumiFileList[0]['lumi_done_time']

            # calculate lumi size and new total size and file count
            lumiSizeTotal = 0
            for mergeFile in lumiFileList:
                lumiSizeTotal += mergeFile['filesize']

            newSizeTotal = jobSizeTotal + lumiSizeTotal
            newFileCount = len(jobFileList) + len(lumiFileList)

            # first lumi, if not old enough bail out
            if len(jobFileList) == 0:
                if lumiAge > timeout:
                    jobFileList.extend(lumiFileList)
                    lastLumi = lumi
                    jobSizeTotal = lumiSizeTotal
                else:
                    break
            # if timeout 0, just mergepack lumi by lumi
            elif timeout == 0:
                self.createMergeJob(jobFileList)
                jobFileList = lumiFileList
                lastLumi = lumi
                jobSizeTotal = lumiSizeTotal
            # observe strict lumi order and sequence broken => mergepack
            # triggers new age check on next out of sequence lumi
            # bail if not old enough
            elif lumi != lastLumi + 1:
                self.createMergeJob(jobFileList)
                if lumiAge > timeout:
                    jobFileList = lumiFileList
                    lastLumi = lumi
                    jobSizeTotal = lumiSizeTotal
                else:
                    jobFileList = []
                    break
            # if below limits just add to mergepack job
            elif newFileCount <= self.maxInputFiles and \
                     ( newSizeTotal <= self.maxInputSize or datatier == "DQM" ):
                jobFileList.extend(lumiFileList)
                lastLumi = lumi
                jobSizeTotal = newSizeTotal
            # over limits => mergepack
            else:
                self.createMergeJob(jobFileList)
                jobFileList = lumiFileList
                lastLumi = lumi
                jobSizeTotal = lumiSizeTotal

        # sequential leftovers that are old enough
        if len(jobFileList) > 0:
            self.createMergeJob(jobFileList)

        return


    def createMergeJob(self, jobFileList):
        """
        _createMergeJob_

        Create an express merge job from list of files

        """
        if not self.createdGroup:
            self.newGroup()
            self.createdGroup = True

        self.newJob(name = makeUUID())
        self.currentJob['mask'].setMaxAndSkipEvents(-1, 0)

        for f in jobFileList:
            self.currentJob.addFile( File(id = f['id']) )
