#!/usr/bin/env python
"""
_RepackerInjectorComponent_

Component for generating Repacker jobsspecs
"""

__version__ = "$Revision: 1.92 $"
__revision__ = "$Id: RepackerInjectorComponent.py,v 1.92 2010/07/21 09:16:18 hufnagel Exp $"

import os
import shutil
import logging

from JobQueue import JobQueueDB

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

from T0.WorkflowFactory.RepackWorkflow import RepackWorkflow
from T0.WorkflowFactory.ConversionWorkflow import ConversionWorkflow

from T0.JobFactory.StreamerJob import StreamerJob

from T0.RepackerInjector.RepackerInjectorAPI import getNewRepackJobs

from T0.RepackerInjector.RepackerInjectorAPI import defineConversionJobs
from T0.RepackerInjector.RepackerInjectorAPI import defineRepackJobs
from T0.RepackerInjector.RepackerInjectorAPI import defineStrictRepackJobs

from T0.RepackerInjector.RepackerInjectorAPI import updateJobs
from T0.RepackerInjector.RepackerInjectorAPI import registerJobs


class RepackerInjectorComponent(GenericTier0Component):
    """
    _RepackerInjectorComponent_

    Component for generating Repacker jobspecs.
    """
    def __init__(self, **args):
        GenericTier0Component.__init__(self, **args) 
        
        self.scramArch = self.lookupParameter("scramArch", "slc4_ia32_gcc345")
        self.cmsPath = self.lookupParameter("cmsPath", "/uscmst1/prod/sw/cms/")
        self.pollInterval = self.lookupParameter("PollInterval", "00:01:00")
        self.jobCacheDir = self.lookupParameter("JobCacheDir",
                                                self.args["ComponentDir"])
        self.workflowSpecQueue = {}

        # For the time being, submit all the jobs as priority 2.
        self.jobQueuePriority = 2

        if self.lookupParameter("UseLazyDownload", "Off") == "On":
            self.useLazyDownload = True
        else:
            self.useLazyDownload = False

        # max file size for EDM RAW, used for output splitting
        self.maxSize = int(self.lookupParameter("MaxFileSize", 10 * 1024 * 1024))

        if self.lookupParameter("SimpleRepacking", "Off") == "On":
            self.simpleRepacking = True
        else:
            self.simpleRepacking = False

        # only used for simple repacking
        self.maxStreamerSizeSingleLumi = int(self.lookupParameter('MaxStreamerSizeSingleLumi', 20*1024*1024*1024))
        self.maxStreamerEventsSingleLumi = int(self.lookupParameter('MaxStreamerEventsSingleLumi', 500000))
        self.maxStreamerSizeMultiLumi = int(self.lookupParameter('MaxStreamerSizeMultiLumi', 6*1024*1024*1024))
        self.maxStreamerEventsMultiLumi = int(self.lookupParameter('MaxStreamerEventsMultiLumi', 500000))
        self.maxStreamerCount = int(self.lookupParameter('MaxStreamerCount', 1000))
        
        if self.lookupParameter("StrictLumiOrder", "Off") == "On":
            self.strictLumiOrder = True
        else:
            self.strictLumiOrder = False

        if self.lookupParameter("SupportConversion", "Off") == "On":
            self.supportConversion = True
        else:
            self.supportConversion = False

        logging.info("RepackerInjector Component Started")


    def __call__(self, message, payload):
        """
        _operator(message, payload)_

        Method called in response to this component receiving a message

        """
        # always have a debug on/off switch
        if message == "RepackerInjector:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif message == "RepackerInjector:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif message == "RepackerInjector:Poll":
            self.poll()
        elif message == "CleanUp":
            self.cleanupRun(payload)

        return


    def cleanupRun(self, runNumber):
        """
        _cleanupRun_

        Remove all workflows and jobspecs for a particular run from disk.
        Also, remove any workflows for the run from memory.
        """
        try:
            runNumber = int(runNumber)
        except ValueError:
            logging.error("Run number is not valid: %s" % runNumber)
            return

        logging.debug("Cleaning up run %s" % runNumber)
        
        # remove workflow from queue
        if runNumber in self.workflowSpecQueue.keys():
            del self.workflowSpecQueue[runNumber]
            
        workflowDir = self.jobCacheDir + "/workflows/Run%s" % runNumber
        jobSpecDir = self.jobCacheDir + "/jobspecs/Run%s" % runNumber
        
        shutil.rmtree(workflowDir,ignore_errors=True)
        shutil.rmtree(jobSpecDir,ignore_errors=True)


    def createWorkflow(self, runNumber, stream_id, stream):
        """
        create the workflow template and publish message for a given run

        """
        # no need to check return value, already did this earlier
        runConfig = self.getRunConfig(runNumber)

        logging.debug("Create new workflow for run %d and stream %s" % (runNumber, stream))

        processingStyle = runConfig.getProcessingStyle(stream_id)

        repackConfig = runConfig.getRepackConfiguration(stream_id)
        if repackConfig == None:
            logging.error("No repack configuration for run %s and stream %s" % (runNumber,
                                                                                stream))
            return None

        if processingStyle == 'Bulk':

            cmsswVer = repackConfig.CMSSWVersion

            outModuleInfoList = runConfig.getOutputModuleInfoList("Repacker", stream_id, "RAW",
                                                                  splitIntoDatasets = True)
            if outModuleInfoList == None:
                logging.error("No output module list for run %s and stream %s" % (runNumber, stream))
                return None

            workflow = RepackWorkflow(runNumber,
                                      cmsswVer,
                                      self.cmsPath,
                                      self.scramArch,
                                      *outModuleInfoList)

            workflowSpec = workflow.makeWorkflow("Repack-Run%s-Stream%s" % (runNumber, stream),
                                                 self.useLazyDownload, self.maxSize,
                                                 production = self.productionMode)
            if workflowSpec == None:
                return None

        elif processingStyle == 'Convert' or processingStyle == 'RegisterAndConvert':

            cmsswVer = repackConfig.CMSSWVersion

            outModuleInfo = runConfig.getConversionOutputModuleInfo(stream_id)

            if outModuleInfo == None:
                logging.error("No output module for run %s and stream %s" % (runNumber, stream))
                return None

            workflow = ConversionWorkflow(runNumber,
                                          cmsswVer,
                                          self.cmsPath,
                                          self.scramArch,
                                          *[outModuleInfo])

            workflowSpec = workflow.makeWorkflow("Repack-Run%s-Stream%s" % (runNumber, stream),
                                                 self.useLazyDownload,
                                                 production = self.productionMode)
            if workflowSpec == None:
                return None

        else:

            logging.error("Unknown processing style %s for run %d and stream %s" %
                          (processingStyle, runNumber, stream))
            return None

        workflowDir = os.path.join(self.jobCacheDir, "workflows",
                                       "Run%s" % runNumber)
        if not os.path.isdir(workflowDir):
            os.makedirs(workflowDir)

        workflowSpecFile = os.path.join(workflowDir,
                                        "Repack-Run%s-Stream%s-workflow.xml" % (runNumber, stream))
        workflowSpec.save(workflowSpecFile)

        logging.debug("Workflow file saved as %s" % workflowSpecFile)

        self.ms.publish("NewWorkflow", workflowSpecFile)
        self.ms.commit()

        return workflowSpec


    def createJobs(self, jobEntityDict, newlyDefinedJobs):
        """
        _createJobs_

        main routine called in polling cycle

        process jobs by run and stream

        """
        createdJobs = []
        createdJobSpecs = []
        for run in sorted(jobEntityDict.keys()):

            # check that this run has a RunConfig record
            runConfig = self.getRunConfig(run)
            if runConfig == None:
                logging.error("Could not retrieve run config for run %s" % run)
                continue

            jobSpecBaseDir = os.path.join(self.jobCacheDir,
                                          "jobspecs",
                                          "Run%d" % run)

            for stream_id in sorted(jobEntityDict[run].keys()):

                workflowSpecsByRun = self.workflowSpecQueue.get(run, None)
                if workflowSpecsByRun == None:
                    workflowSpecsByRun = {}
                    self.workflowSpecQueue[run] = workflowSpecsByRun

                workflowSpec = workflowSpecsByRun.get(stream_id, None)

                if workflowSpec == None:
                    stream = runConfig.getStreamName(stream_id)

                    workflowSpec = self.createWorkflow(run, stream_id, stream)
                    if workflowSpec == None:
                        logging.error("Can't create workflow for run %d and stream %s"
                                      % (run, stream))
                        continue
                    else:
                        self.workflowSpecQueue[run][stream_id] = workflowSpec

                # reuse job factory for all jobs in a run and stream
                jobFactory = StreamerJob(workflowSpec)
                for jobEntity in jobEntityDict[run][stream_id]:
                    jobSpec = jobFactory("Repack", jobEntity, jobSpecBaseDir)
                    createdJobs.append(jobEntity)
                    createdJobSpecs.append(jobSpec)

        if len(createdJobs) > 0:

            try:
                self.paDBConn.beginTransaction()

                for jobSpec in createdJobSpecs:
                    JobQueueDB.insertJobSpec(self.paDBConn.getDBInterface(),
                                             self.paDBConn.dbTransaction.conn,
                                             jobSpec.parameters["JobName"],
                                             jobSpec.parameters['JobSpecFile'],
                                             jobSpec.parameters["JobType"],
                                             jobSpec.payload.workflow,
                                             self.jobQueuePriority,
                                             siteName = None,
                                             status = "new")

                self.t0astDBConn.beginTransaction()

                if newlyDefinedJobs:
                    registerJobs(self.t0astDBConn, createdJobs)
                else:
                    updateJobs(self.t0astDBConn, createdJobs)

                self.paDBConn.commit()
                self.t0astDBConn.commit()
                logging.info("Created and queued %d new jobs" % len(createdJobs))
            except StandardError, ex:
                self.t0astDBConn.rollback()
                self.paDBConn.rollback()
                logging.error("Failed to create jobs : %s\n" % str(ex))

        return len(createdJobs)


    def poll(self):
        """
        _poll_

        Single cycle invocation of job creation

        """
        logging.info("Job creation poll invoked")

        # handle repack jobs (simple or normal mode)
        if self.simpleRepacking:

            if self.strictLumiOrder:
                jobEntityDict = defineStrictRepackJobs(self.t0astDBConn,
                                                       self.maxStreamerSizeSingleLumi,
                                                       self.maxStreamerEventsSingleLumi,
                                                       self.maxStreamerSizeMultiLumi,
                                                       self.maxStreamerEventsMultiLumi,
                                                       self.maxStreamerCount)
            else:
                jobEntityDict = defineRepackJobs(self.t0astDBConn,
                                                 self.maxStreamerSizeSingleLumi,
                                                 self.maxStreamerEventsSingleLumi,
                                                 self.maxStreamerSizeMultiLumi,
                                                 self.maxStreamerEventsMultiLumi,
                                                 self.maxStreamerCount)
            logging.info("Creating %d simple repack jobs" % len(jobEntityDict.keys()))
            self.createJobs(jobEntityDict, True)
        else:
            jobEntityDict = getNewRepackJobs(self.t0astDBConn)
            logging.info("Creating %d normal repack jobs" % len(jobEntityDict.keys()))
            self.createJobs(jobEntityDict, False)

        # last handle conversion jobs
        if self.supportConversion:
            jobEntityDict = defineConversionJobs(self.t0astDBConn)
            self.createJobs(jobEntityDict, True)
        
        self.ms.publishUnique("RepackerInjector:Poll", "", self.pollInterval)
        self.ms.commit()

        return
    

    def startComponent(self):
        """
        _startComponent_

        Start up the component and define the messages that it subscribes to.
        """
        self.ms.subscribeTo("RepackerInjector:StartDebug")
        self.ms.subscribeTo("RepackerInjector:EndDebug")
        self.ms.subscribeTo("RepackerInjector:Poll")

        # published by JobScheduler Component
        self.ms.subscribeTo("NewRepackJobs")
        self.ms.subscribeTo("CleanUp")
        
        self.connectT0AST()
        self.connectPADB()

        self.ms.publishUnique("RepackerInjector:Poll", "", "00:01:00")
        self.ms.commit()
        
        # wait for messages
        self.messageLoop()
