#!/usr/bin/env python
"""
_RepackerTesterComponent_

Simple component to inject bogus data into T0AST which can be used to
test other components.
"""

__revision__ = "$Id: RepackerTesterComponent.py,v 1.70 2009/12/05 17:43:01 hufnagel Exp $"
__version__ = "$Revision: 1.70 $"

import logging
import random
import time
import sys
import threading

from ProdCommon.MCPayloads.UUID import makeUUID

from WMCore.WMBS.File import File
from WMCore.WMBS.Fileset import Fileset
from WMCore.WMBS.Workflow import Workflow
from WMCore.WMBS.Job import Job
from WMCore.WMBS.JobGroup import JobGroup
from WMCore.WMBS.Subscription import Subscription
from WMCore.DAOFactory import DAOFactory
from WMCore.DataStructs.Run import Run

from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Reader import ListRunConfig
from T0.State.Database.Writer import InsertDataset
from T0.State.Database.Writer import InsertJob
from T0.State.Database.Writer import InsertRunConfig
from T0.State.Database.Writer import InsertTriggerSegment

from T0.GenericTier0.GenericTier0Component import GenericTier0Component

class RepackerTesterComponent(GenericTier0Component):
    """
    _RepackerTesterComponent_

    Simple component to inject bogus data into T0AST which can be used to
    test other components.
    """
    def __init__(self, **args):
        """
        ___init___

        Initialize class data members and verify any config options.
        """
        GenericTier0Component.__init__(self, **args)
        logging.info("Repacker Tester Component Started")

        # How often this component should inject data into T0AST.
        self.injectInterval = self.lookupParameter("injectInterval", "00:01:00")

        # What kind of data will be injected into T0AST.  Currently, the only
        # option is Merger which will inject repacked files into the repacked
        # table.
        self.injectMode = self.lookupParameter("injectMode", "Merger")        

        # Keep track of the last RunNumber we injected into the database
        # because there isn't a sequence to keep track of it for us...
        self.lastInjectedRun = 0

        # Keep track of the last Lumi Section we injected into the database
        # because there isn't a sequence to keep track of it for us...
        self.lastInjectedLumiSection = -1        

        # Keep track of the last stream we injected into the database
        # because there isn't a sequence to keep track of it for us...
        self.lastInjectedStream = -1

        # Keep track of the datasets that we inject.
        self.lastInjectedDataset = 0
        
        # Keep track of active runs so we don't have to query T0AST.
        self.activeRuns = []

        # Keep track of active streams so that we don't have to query T0AST.
        self.activeStreams = ["A", "HLTDEBUG", "Calibration"]
        self.expressStreams = ["Express"]

        # A list of the algorithms that are supported by the scheduler and
        # the injector.
        self.algorithmList = ["split", "accumulate"]

        # List of different types of streamers that can be injected.
        self.streamerTypes = ["repack", "express", "pixdmp"]
        
        # Keep track of the event number so that we can easily set the
        # first_event and last_event fields for wmbs files.
        self.firstEvent = 0

        # Name of the storage node that the RepackerTester creates for skimming
        # tests.
        self.storageNode = None
        return

    def __call__(self, event, payload):
        """
        _operator(message, payload)_

        Method called in response to this component recieving a message.
        """
        logging.debug("Received a message: %s(%s)" % (event, payload))

        # always have a debug on/off switch
        if event == "RepackerTester:StartDebug":
            logging.getLogger().setLevel(logging.DEBUG)
        elif event == "RepackerTester:EndDebug":
            logging.getLogger().setLevel(logging.INFO)
        elif event == "RepackerTester:Purge":
            self.purgeT0AST()
        elif event == "RepackerTester:EndRun":
            self.endRun(payload)
        elif event == "RepackerTester:ContinueRun":
            self.continueRun(payload)
        # There are several modes that the RepackerTester can run in:
        #   Merger: Insert rows into the repacked table that can be merged
        #           into larger files.
        #   Injector: Insert rows into all the necessary tables so repack job
        #             definitions can be created and used by the
        #             RepackerInjector.
        #   Scheduler: Similar to Injector mode, but do not create the repack
        #              job defs.
        #   MergeInject: Run both Merger and Injector modes.
        #   KitchenSink: A combination of Scheduler and Merger mode, meant to
        #                use with the whole system.
        #   PromptReco: Used to test the PromptRecoInjector.
        #   AlcaReco: Used to test the AlcaSkimInjector.
        #   Export: Used to test the export system.
        #   Skim: Used to test the Tier1 Skimming system.
        #   DQM: Inject RECO files into the DQM fileset.
        elif event == "RepackerTester:Inject":
            self.t0astDBConn.beginTransaction()
            if self.injectMode == "Merger":
                self.injectMerger()
                self.injectMerger()
                self.injectMerger()
                self.injectMerger()
                self.injectMerger()                
            elif self.injectMode == "Injector":
                self.injectInjector()
            elif self.injectMode == "Scheduler":
                self.injectScheduler()
            elif self.injectMode == "MergeInject":
                self.injectInjector()
                self.injectMerger()
            elif self.injectMode == "KitchenSink":
                self.injectScheduler()                
                self.injectMerger()
                self.injectPromptReco()
            elif self.injectMode == "PromptReco":
                self.injectPromptReco()
            elif self.injectMode == "AlcaReco":
                self.injectAlcaReco()                
            elif self.injectMode == "Export":
                self.injectExport()
            elif self.injectMode == "Skim":
                self.injectSkim()
                self.injectSkim()
                self.injectSkim()
                self.injectSkim()
                self.injectSkim()
            elif self.injectMode == "DQM":
                self.injectDQM()
            else:
                logging.error("Unknown inject mode: %s" % self.injectMode)

            self.t0astDBConn.commit()
            self.ms.publishUnique("RepackerTester:Inject", "",
                                  self.injectInterval)
            self.ms.commit()            
        
        return

    def queryLumiSectionsForRun(self, runNumber):
        """
        _queryLumiSectionsForRun_

        Query T0AST to determine which lumi sections are associated with a
        particular run number and return a list containing those lumi sections
        IDs.
        """
        lumiQuery = "SELECT LUMI_ID FROM lumi_section "
        lumiQuery += "WHERE RUN_ID = %d" % runNumber

        self.t0astDBConn.execute(lumiQuery)
        lumiRows = self.t0astDBConn.fetchall()        

        lumiIDs = []
        for lumiRow in lumiRows:
            lumiIDs.append(int(lumiRow[0]))

        return lumiIDs

    def injectExpressStream(self):
        """
        inject express stream in stream table
        """
        sqlQuery = """ INSERT INTO stream (id, name) VALUES (stream_SEQ.nextval, 
                       :p_1) """               
        binds = []
        for stream in self.expressStreams:
            param = {}
            param["p_1"] = stream
            binds.append(param)
        
        self.t0astDBConn.execute(sqlQuery, binds)
        
    def injectRun(self):
        """
        _injectRun_

        Inject a run into T0AST.  We only need to fill in the run_id,
        run_status, start_time, end_time, version and hltkey fields.  The rest
        will be filled in by RunConfig.
        """
        self.lastInjectedRun += 1
        self.activeRuns.append(self.lastInjectedRun)

        hltKey = "/cdaq/cosmic/commissioning2008/CRAFT/v4.0/HLT/V1"

        InsertRunConfig.insertCMSSWVersion(self.t0astDBConn, "CMSSW_2_2_6")
        InsertRunConfig.insertCMSSWVersion(self.t0astDBConn, "CMSSW_2_2_7")
        InsertRunConfig.insertCMSSWVersion(self.t0astDBConn, "CMSSW_2_2_8")        

        sqlQuery = """INSERT INTO run (RUN_ID, START_TIME, END_TIME, HLTKEY,
                      RUN_VERSION, REPACK_VERSION, EXPRESS_VERSION, RUN_STATUS)
                      VALUES (:runid, :begin, :finish, :hltkey,
                      (SELECT id FROM cmssw_version WHERE name = :runver),
                      (SELECT id FROM cmssw_version WHERE name = :repackver),
                      (SELECT id FROM cmssw_version WHERE name = :expressver),
                      (SELECT id FROM run_status WHERE STATUS = :status))"""
        
        dataToInsert = {}
        dataToInsert["runid"] = self.lastInjectedRun
        dataToInsert["begin"] = int(time.time())
        dataToInsert["finish"] = int(time.time()) + 3600
        dataToInsert["hltkey"] = hltKey
        dataToInsert["runver"] = "CMSSW_2_2_6"
        dataToInsert["repackver"] = "CMSSW_2_2_7"
        dataToInsert["expressver"] = "CMSSW_2_2_8"
        dataToInsert["status"] = "Active"
        self.t0astDBConn.execute(sqlQuery, dataToInsert)

        runConfig = self.getRunConfig(self.lastInjectedRun)
        return self.lastInjectedRun

    def endRun(self, runNumber):
        """
        _endRun_

        Remove the run from the activeRuns list so we stop injecting data into
        it.
        """
        try:
            runNumber = int(runNumber)
        except ValueError:
            logging.error("Run number is not valid: %s" % runNumber)
            return
                                                
        if runNumber in self.activeRuns:
            self.activeRuns.remove(runNumber)
        else:
            logging.debug("Run %s is not active." % runNumber)
            return

        logging.debug("Removed run %s from the active list." % runNumber)

    def continueRun(self, runNumber):
        """
        _continueRun_

        Add a run number to the active list so that we start generating data
        for it again.  Note, the run must already exist in the run table.
        """
        try:
            runNumber = int(runNumber)
        except ValueError:
            logging.error("Run number is not valid: %s" % runNumber)
            return
                                                
        if runNumber not in self.activeRuns:
            self.activeRuns.append(runNumber)
        else:
            logging.debug("Run %s is already active." % runNumber)
            return

        logging.debug("Added run %s to the active list." % runNumber)        

    def injectLumiSection(self, runNumber):
        """
        _injectLumiSection_

        Inject a Lumi Section into T0AST for a particular run number.
        """
        sqlQuery = """INSERT INTO lumi_section (LUMI_ID, RUN_ID, START_TIME)
                      VALUES (:p_1, :p_2, :p_3)"""
        
        self.lastInjectedLumiSection += 1
        
        dataToInsert = {}        
        dataToInsert["p_1"] = self.lastInjectedLumiSection
        dataToInsert["p_2"] = runNumber
        dataToInsert["p_3"] = int(time.time())
        self.t0astDBConn.execute(sqlQuery, dataToInsert)

        return self.lastInjectedLumiSection

    def injectStreamer(self, lumiID, runNumber):
        """
        _injectStreamer_

        Inject a streamer into T0AST for a particular run number and lumi
        section.  Return the streamer ID.
        """
        self.t0astDBConn.execute("SELECT streamer_SEQ.nextval from dual")
        streamerID = int(self.t0astDBConn.fetchall()[0][0])

        sqlQuery = """INSERT INTO streamer (STREAMER_ID, RUN_ID, LUMI_ID,
                      INSERT_TIME, FILESIZE, EVENTS, LFN, EXPORTABLE,
                      STREAM_ID, INDEXPFN, INDEXPFNBACKUP) SELECT
                      :STREAMER_ID, :RUN_ID, :LUMI_ID, :INSERT_TIME, :FILESIZE,
                      :EVENTS, :LFN,
                      :EXPORTABLE,
                      (SELECT id FROM stream WHERE name = :STREAM),
                      :INDEXPFN, :INDEXPFNBACKUP FROM DUAL"""

        dataToInsert = {}
        dataToInsert["STREAMER_ID"] = streamerID
        dataToInsert["RUN_ID"] = runNumber
        dataToInsert["LUMI_ID"] = lumiID
        dataToInsert["INSERT_TIME"] = int(time.time())        
        dataToInsert["FILESIZE"] = random.randrange(1024, 1248576, 1024)
        dataToInsert["EVENTS"] = random.randrange(1, 1024, 1)
        dataToInsert["LFN"] = makeUUID()
        dataToInsert["EXPORTABLE"] = 0
        dataToInsert["STREAM"] = random.choice(self.activeStreams)
        dataToInsert["INDEXPFN"] = makeUUID()
        dataToInsert["INDEXPFNBACKUP"] = makeUUID()
        self.t0astDBConn.execute(sqlQuery, dataToInsert)

        return streamerID                        

    def injectTriggerSegment(self, primaryDatasetID, streamerID, lumiID, runID):
        """
        _injectTriggerSegment_

        Inject a trigger segment into T0AST given a primary dataset ID,
        streamer ID, lumi ID and run number.
        """
        triggerSegment = {}
        triggerSegment["dataset_id"] = primaryDatasetID
        triggerSegment["streamer_id"] = streamerID
        triggerSegment["lumi_id"] = lumiID
        triggerSegment["run_id"] = runID
        triggerSegment["segment_size"] = random.randrange(1024, 1248576, 1024)
        triggerSegment["status"] = 1

        InsertTriggerSegment.insertTriggerSegments(self.t0astDBConn,
                                                   triggerSegment)
        return

    def assocRepackedStreamer(self, streamerID, repackedID):
        """
        _assocRepackedStreamer_

        Associate a repacked file with a particular streamer by adding a
        row to the repack_streamer_assoc table.
        """
        sqlQuery = """INSERT INTO repack_streamer_assoc (STREAMER_ID,
                      REPACKED_ID) VALUES (:p_1, :p_2)"""
        
        dataToInsert = {}
        dataToInsert["p_1"] = streamerID
        dataToInsert["p_2"] = repackedID
        self.t0astDBConn.execute(sqlQuery, dataToInsert)

    def injectWMBSFile(self, fileset = None, primaryDatasetName = None,
                       processedDatasetName = None, dataTierName = None,
                       lumiSection = None):
        """
        _injectWMBSFile_

        Inject a file into WMBS and add it to a specific fileset.  This will
        also associate the file to a particular primary dataset,
        processed dataset, data tier and lumi section.  If any of the parameters
        are not defined random ones will be chosen.  This method will return the
        ID of the file that was created.
        """
        events = random.randrange(20000, 40000, 1000)

        if random.randrange(0, 10, 1) == 1:
            events = 0
            
        wmbsFile = File(lfn = makeUUID(), size = random.randrange(1024, 1248576, 1024),
                        events = events, cksum = 12345, first_event = self.firstEvent,
                        last_event = self.firstEvent + events, locations = "se1.cern.ch")

        self.firstEvent += events

        randomRun = random.choice(self.activeRuns)

        if lumiSection == None:
            chosenLumi = random.sample(self.queryLumiSectionsForRun(randomRun), 1)
        else:
            chosenLumi = [lumiSection]

        wmbsRun = Run(randomRun, *chosenLumi)
        wmbsFile.addRun(wmbsRun)
        wmbsFile.create()

        runConfig = self.getRunConfig(randomRun)
        if runConfig == None:
            logging.error("Couldn't get RunConfig for run %s" % randomRun)
            return None

        primaryDatasets = runConfig.getPrimaryDatasets()
        if primaryDatasets == None:
            logging.error("Couldn't get primary datasets for run %s" % randomRun)
            return None

        processedDatasets = ListDatasets.listProcessedDatasets(self.t0astDBConn)
        dataTiers = ListDatasets.listDataTiers(self.t0astDBConn)

        if primaryDatasetName == None:
            chosenPrimaryDataset = random.choice(primaryDatasets.keys())
        else:
            for primaryDatasetID in primaryDatasets.keys():
                if primaryDatasets[primaryDatasetID] == primaryDatasetName:
                    chosenPrimaryDataset = primaryDatasetID
                    break

        if processedDatasetName == None:
            chosenProcessedDataset = random.choice(processedDatasets.keys())
        else:
            for processedDatasetID in processedDatasets.keys():
                if processedDatasets[processedDatasetID] == processedDatasetName:
                    chosenProcessedDataset = processedDatasetID
                    break
                
        if dataTierName == None:
            chosenDataTier = random.choice(dataTiers.keys())
        else:
            for dataTierID in dataTiers.keys():
                if dataTiers[dataTierID] == dataTierName:
                    chosenDataTier = dataTierID
                    break
            
        InsertDataset.assocWMBSFileDataset(self.t0astDBConn,
                                           chosenPrimaryDataset,
                                           chosenProcessedDataset,
                                           chosenDataTier, wmbsFile["id"])

        if fileset != None:
            logging.debug("FILESET: %s" % fileset)
            wmbsFileset = Fileset(name = fileset)
            wmbsFileset.load()
            wmbsFileset.addFile(wmbsFile)
            wmbsFileset.commit()
            
        return wmbsFile["id"]

    def injectBlock(self, datasetPathID, fileIDs):
        """
        _injectBlock_

        Inject a block into T0AST for a given dataset path and containing the
        given fileIDs.  This method will insert data into the block,
        block_runs_assoc, wmbs_file_block_assoc and phedex_subscription_made
        tables.  Blocks will be injected with a status of "Migrated".
        """
        self.t0astDBConn.execute("SELECT block_SEQ.nextval from dual")
        blockID = self.t0astDBConn.fetchall()[0][0]
        
        sqlQuery = """INSERT INTO block (id, dataset_path_id, name, block_size,
                      file_count, status, migrate_status) VALUES (:p_1, :p_2, :p_3,
                      :p_4, :p_5,
                      (SELECT id FROM block_status WHERE status = :p_6),
                      (SELECT id FROM block_migrate_status WHERE status = :p_7))
                      """
        bindVars = {}
        bindVars["p_1"] = blockID
        bindVars["p_2"] = datasetPathID

        datasetNames = ListDatasets.listDatasetNamesFromID(self.t0astDBConn,
                                                           datasetPathID)
        blockName = "/%s/%s/%s#%s" % (datasetNames["PRIMARY"],
                                      datasetNames["PROCESSED"],
                                      datasetNames["TIER"],
                                      makeUUID())
        bindVars["p_3"] = blockName
        
        bindVars["p_4"] = 1
        bindVars["p_5"] = 1
        bindVars["p_6"] = "Exported"
        bindVars["p_7"] = "Migrated"
        self.t0astDBConn.execute(sqlQuery, bindVars)

        sqlQuery = """INSERT INTO block_run_assoc (run_id, block_id)
                      VALUES (:p_1, :p_2)"""
        bindVars = {"p_1": random.choice(self.activeRuns), "p_2": blockID}
        self.t0astDBConn.execute(sqlQuery, bindVars)

        bindVars = []
        for fileID in fileIDs:
            bindVar = {"BLOCK_ID": blockID, "FILE_ID": fileID}
            bindVars.append(bindVar)

        sqlQuery = """INSERT INTO wmbs_file_block_assoc (file_id, block_id)
                      VALUES (:FILE_ID, :BLOCK_ID)"""

        self.t0astDBConn.execute(sqlQuery, bindVars)

        sqlQuery = """INSERT INTO phedex_subscription_made
                        (dataset_path_id, node_id, requester_id)
                      SELECT :p_1, :p_2, :p_3 FROM DUAL WHERE NOT EXISTS
                         (SELECT * FROM phedex_subscription_made
                           WHERE dataset_path_id = :p_1 AND node_id = :p_2)"""

        bindVars = {"p_1": datasetPathID, "p_2": self.storageNode, "p_3": 0}
        self.t0astDBConn.execute(sqlQuery, bindVars)
        
        return

    def injectInjector(self):
        """
        _injectInjector_

        Inject lumi sections, streamers, trigger segments and repack job
        definitions that the RepackerInjecter can use to create repack jobs.
        """
        if len(self.activeRuns) == 0:
            return
        
        runNumber = self.activeRuns[random.randrange(len(self.activeRuns))]
        if ListRunConfig.runExists(self.t0astDBConn, runNumber) == False:
            return

        runConfig = self.getRunConfig(runNumber)
        if runConfig == None:
            logging.error("Could not retrieve runconfig for run %s" % runNumber)
            return

        primaryDatasets = runConfig.getPrimaryDatasets()
        if primaryDatasets == None:
            logging.error("Could not retrieve primary datasets for run %s" %
                          runNumber)
            return

        primaryDatasetIDs = primaryDatasets.keys()
        
        # inject a lumi section with the run id, get the lumi_id back
        lumiID = self.injectLumiSection(runNumber)
        
        # inject two streamers with the lumi id and run id.
        streamerIDs = []
        streamerIDs.append(self.injectStreamer(lumiID, runNumber))
        streamerIDs.append(self.injectStreamer(lumiID, runNumber))
        
        # inject (num streamers * num datasets) trig segments
        for primaryDatasetID in primaryDatasetIDs:
            for streamerID in streamerIDs:
                self.injectTriggerSegment(primaryDatasetID, streamerID, lumiID,
                                          runNumber)
        
        # inject num trig segments repack job defs, save the ids
        repackJobDefIDs = []
        for i in range(len(primaryDatasetIDs) * len(streamerIDs)):
            jobID = InsertJob.insertJobDef(self.t0astDBConn, "Repack", "New")
            repackJobDefIDs.append(jobID)
            
        # Associate the job with a dataset and streamer.
        sqlQuery = """INSERT INTO repack_job_streamer_assoc
                      (JOB_ID, STREAMER_ID)
                      VALUES (:p_1, :p_2)"""
        bindVars = []
        count = 0
        for repackedJobDefID in repackJobDefIDs:
            dataToInsert = {}
            dataToInsert["p_1"] = repackedJobDefID            
            dataToInsert["p_2"] = streamerIDs[count % len(streamerIDs)]
            bindVars.append(dataToInsert)
            count += 1

        self.t0astDBConn.execute(sqlQuery, bindVars)

    def injectScheduler(self):
        """
        _injectScheduler_

        Inject streamers, lumi sections and trigger segments that the
        RepackerScheduler can use to create repack jobs.
        """
        if len(self.activeRuns) == 0:
            return
        
        runNumber = self.activeRuns[random.randrange(len(self.activeRuns))]
        if ListRunConfig.runExists(self.t0astDBConn, runNumber) == False:
            return

        runConfig = self.getRunConfig(runNumber)
        if runConfig == None:
            logging.error("Could not retrieve runconfig for run %s" % runNumber)
            return

        primaryDatasets = runConfig.getPrimaryDatasets()
        if primaryDatasets == None:
            logging.error("Could not retrieve primary datasets for run %s" %
                          runNumber)
            return

        primaryDatasetIDs = primaryDatasets.keys()
        
        # inject a lumi section with the run id, get the lumi_id back
        lumiID = self.injectLumiSection(runNumber)
        
        # inject two streamers with the lumi id and run id.
        streamerIDs = []
        streamerIDs.append(self.injectStreamer(lumiID, runNumber))
        streamerIDs.append(self.injectStreamer(lumiID, runNumber))
        
        # inject (num streamers * num datasets) trig segments
        for primaryDatasetID in primaryDatasetIDs:
            for streamerID in streamerIDs:
                self.injectTriggerSegment(primaryDatasetID, streamerID,
                                          lumiID, runNumber)
        
        return
        
    def injectMerger(self):
        """
        _injectMerger_

        Inject repacked files into T0AST that can be used to test the
        Tier0Merger.  Also inject some completed WMBS job groups that
        need to be merged.
        """
        if len(self.activeRuns) == 0:
            return
        
        filesToCreate = random.randrange(0, 10, 1)        

        logging.debug("injectMerger(): Going to inject %d files" % \
                                       filesToCreate)

        if filesToCreate == 0:
            return

        for i in range(filesToCreate):
            self.injectWMBSFile(fileset = "Mergeable")

        # Create an input file, save the ID
        randomRun = random.choice(self.activeRuns)
        chosenLumi = random.choice(self.queryLumiSectionsForRun(randomRun))
        inputFile = self.injectWMBSFile(fileset = "Reconstructable", lumiSection = chosenLumi)

        # Find the reco subscription
        wmbsRecoFileset = Fileset(name = "Reconstructable")
        wmbsRecoFileset.load()
        wmbsRecoWorkflow = Workflow(name = "FileReconstruction", task = "Bogus")
        wmbsRecoWorkflow.load()
        wmbsRecoSubscription = Subscription(fileset = wmbsRecoFileset,
                                            workflow = wmbsRecoWorkflow)
        wmbsRecoSubscription.load()
        
        # Create JobGroup, use the Reconstructable subscription
        wmbsJobGroup = JobGroup(subscription = wmbsRecoSubscription)
        wmbsJobGroup.create()

        # Determine how many jobs/output files we will have
        totalJobs = random.choice(range(1,10))

        changeStateAction = self.daoFactory(classname = "Jobs.ChangeState")
        
        # Create N jobs, using the file as input, update the masks, mark the jobs as complete
        for i in range(totalJobs):
            wmbsJob = Job()
            wmbsJob.create(group = wmbsJobGroup)
            wmbsJob["state"] = "cleanout"
            changeStateAction.execute(jobs = [wmbsJob])
            
        # Get the output dataset IDs
        outputDataset = ListDatasets.listDatasetNamesForWMBSFile(self.t0astDBConn,
                                                                 inputFile)
        
        # Create N output files, associate them to the output of the jobgroup
        for i in range(totalJobs):
            wmbsOutputFileID = self.injectWMBSFile(fileset = "Mergeable",
                                                 primaryDatasetName = outputDataset["PRIMARY"],
                                                 processedDatasetName = outputDataset["PROCESSED"],
                                                 dataTierName = outputDataset["TIER"],
                                                 lumiSection = chosenLumi)
            wmbsOutputFile = File(id = wmbsOutputFileID)
            wmbsOutputFile.load()
            wmbsJobGroup.output.addFile(wmbsOutputFile)
            
        wmbsJobGroup.output.commit()

        return

    def injectSkim(self):
        """
        _injectSkim_

        Inject WMBS files and add them to blocks so that Tier1 skimming can be
        tested.
        """
        fileID = self.injectWMBSFile(fileset = "Transferable")
        datasetPathID = ListDatasets.listDatasetIDForWMBSFile(self.t0astDBConn,
                                                              fileID)
        datasetNames = ListDatasets.listDatasetNamesFromID(self.t0astDBConn,
                                                           datasetPathID)

        fileIDs = [fileID]
        for i in range(4):
            fileID = self.injectWMBSFile(fileset = "Transferable", 
                                         primaryDatasetName = datasetNames["PRIMARY"],
                                         processedDatasetName = datasetNames["PROCESSED"],
                                         dataTierName = datasetNames["TIER"])
            fileIDs.append(fileID)

        for childID in fileIDs:
            parentID = self.injectWMBSFile(fileset = "Transferable", 
                                           primaryDatasetName = datasetNames["PRIMARY"],
                                           processedDatasetName = datasetNames["PROCESSED"],
                                           dataTierName = datasetNames["TIER"])
            parentFile = File(id = parentID)
            childFile = File(id = childID)
            childFile.load()
            parentFile.addChild(lfn = childFile["lfn"])

        self.injectBlock(datasetPathID, fileIDs)
        return

    def injectPromptReco(self):
        """
        _injectPromptReco_

        Inject repacked files into T0AST that can be used to test the
        PromptRecoInjector.  
        """
        if len(self.activeRuns) == 0:
            return
        
        filesToCreate = random.randrange(0, 10, 1)

        logging.debug("injectPromptReco(): Going to inject %d files" % \
                      filesToCreate)

        if filesToCreate == 0:
            return

        for i in range(filesToCreate):
            self.injectWMBSFile(fileset = "Reconstructable", dataTierName = "RAW")            

        return

    def injectAlcaReco(self):
        """
        _injectAlcaReco_

        Inject reco files into T0AST that can be used to test the
        AlcaSkimInjector.
        """
        if len(self.activeRuns) == 0:
            return
        
        filesToCreate = random.randrange(0, 10, 1)

        logging.debug("injectAlcaReco(): Going to inject %d files" % \
                      filesToCreate)

        if filesToCreate == 0:
            return

        for i in range(filesToCreate):
            self.injectWMBSFile(fileset = "AlcaSkimmable", dataTierName = "RECO")            

        return

    def injectDQM(self):
        """
        _injectDQM_

        Inject reco files into T0AST that can be used to test the
        DQMScheduler.
        """
        if len(self.activeRuns) == 0:
            return
        
        filesToCreate = random.randrange(0, 10, 1)

        logging.debug("injectDQM(): Going to inject %d files" % \
                      filesToCreate)

        if filesToCreate == 0:
            return

        for i in range(filesToCreate):
            self.injectWMBSFile(fileset = "DQMenable", dataTierName = "RECO")            

        return    

    def injectExport(self):
        """
        _injectExport_

        Inject files into T0AST that can be used to test the export system.
        """
        if len(self.activeRuns) == 0:
            return
        
        filesToCreate = random.randrange(0, 10, 1)

        logging.debug("injectExport(): Going to inject %d files" % filesToCreate)

        if filesToCreate == 0:
            return

        for i in range(filesToCreate):
            self.injectWMBSFile(fileset = "DBSUploadable")            

        return
            
    def genericTestSetup(self):
        """
        _genericTestSetup_

        Add three runs and let the RunConfig fill in the rest.

        Also, when in "Merger" mode at five lumi sections for each run that
        is created so repacked/lumi associations can be setup correctly.
        """
        logging.debug("genericTestSetup(): Running...")
        self.t0astDBConn.beginTransaction()

        InsertDataset.insertProcessedDataset(self.t0astDBConn, "bogus-v1")
        #InsertDataset.insertProcessedDataset(self.t0astDBConn, "bogus-v2")
        #InsertDataset.insertProcessedDataset(self.t0astDBConn, "bogus-v3")        
        
        self.injectExpressStream()
        # Add three runs...
        self.injectRun()
        self.injectRun()
        self.injectRun()

        if self.injectMode == "Merger" or self.injectMode == "KitchenSink" or \
               self.injectMode == "PromptReco" or self.injectMode == "Export" or \
               self.injectMode == "Skim" or self.injectMode == "AlcaReco" or \
               self.injectMode == "DQM":
            # We need to create some lumi sections for merge mode so that
            # the repack_lumi_assoc table can be populated.
            for activeRun in self.activeRuns:
                for i in range(5):
                    self.injectLumiSection(activeRun)

        if self.injectMode == "PromptReco":
            # Inject some streamers so that the code that releases jobs
            # can look at the time stamps on the streamers.
            for activeRun in self.activeRuns:
                lumis = self.queryLumiSectionsForRun(activeRun)
                self.injectStreamer(lumis[0], activeRun)            

        if self.injectMode == "Skim":
            self.t0astDBConn.execute("SELECT storage_node_SEQ.nextval from dual")
            self.storageNode = int(self.t0astDBConn.fetchall()[0][0])        
            query = "INSERT INTO storage_node (id, name) VALUES (:p_1, :p_2)"
            bindVars = {"p_1": self.storageNode, "p_2": "se.fnal.gov"}
            self.t0astDBConn.execute(query, bindVars)

        # Add a location to WMBS so we can store our files there.
        locationAction = self.daoFactory(classname = "Locations.New")
        locationAction.execute(siteName = "se1.cern.ch")
        self.t0astDBConn.commit()
        return

    def startComponent(self):
        """
        _startComponent_

        Subscribe to relevant messages, connect to T0AST and then enter the
        message loop.
        """
        self.ms.subscribeTo("RepackerTester:StartDebug")
        self.ms.subscribeTo("RepackerTester:EndDebug")
        self.ms.subscribeTo("RepackerTester:Purge")
        self.ms.subscribeTo("RepackerTester:Inject")
        self.ms.subscribeTo("RepackerTester:EndRun")
        self.ms.subscribeTo("RepackerTester:ContinueRun")
        self.ms.commit()

        self.connectT0AST()

        myThread = threading.currentThread()
        self.daoFactory = DAOFactory(package = "WMCore.WMBS",
                                     logger = logging,
                                     dbinterface = myThread.dbi)

        self.genericTestSetup()

        self.ms.publishUnique("RepackerTester:Inject", "")
        self.ms.commit()

        # wait for messages
        self.messageLoop()
