#!/usr/bin/env python
"""
_ListStreamers_

Functions for retreiving information from the streamer table in T0AST.
"""

__version__ = "$Revision: 1.43 $"
__revision__ = "$Id: ListStreamers.py,v 1.43 2010/07/21 09:15:24 hufnagel Exp $"

import logging
from T0.DataStructs.StreamerFile import StreamerFile

def getRunsWithoutRunConfig(dbConn):
    """
    _getRunsWithoutRunConfig_

    returns list of runs without acqusition era
    (ie. RunConfig not loaded) and unused streamers

    later might not be needed, but it's not expensive
    and a safety check against runs with no data that
    somehow ended up in T0AST

    """
    sqlQuery = """SELECT DISTINCT run.run_id
                  FROM run
                  INNER JOIN streamer ON
                    streamer.run_id = run.run_id AND
                    checkForZeroState(streamer.used) = 0
                  WHERE run.acq_era IS NULL"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    runs = []
    for result in results:
        runs.append(result[0])

    return runs

def getAvailableExpressStreamers(dbConn):
    """
    _getAvailableExpressStreamers_

    returns all unused streamers from streams
    with processing style 'Express'

    """
    sqlQuery = """SELECT streamer.streamer_id, streamer.run_id, streamer.lumi_id,
                         streamer.events, streamer.lfn, streamer.stream_id
                  FROM streamer
                  INNER JOIN run_stream_style_assoc ON
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Express')
                  INNER JOIN run ON
                    run.run_id = streamer.run_id AND
                    run.express_started = 1
                  WHERE checkForZeroState(streamer.used) = 0"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    streamers = []
    for result in results:
        streamerDict = {}
        streamerDict['STREAMER_ID'] = result[0]
        streamerDict['RUN_ID'] = result[1]
        streamerDict['LUMI_ID'] = result[2]
        streamerDict['EVENTS'] = result[3]
        streamerDict['LFN'] = result[4]
        streamerDict['STREAM_ID'] = result[5]
        streamers.append(streamerDict)

    return streamers

def getAvailableRepackStreamers(dbConn):
    """
    _getAvailableRepackStreamers_

    returns all unused streamers from streams
    with processing style 'Bulk' whose
    lumi section is closed

    """
    sqlQuery = """SELECT run_id
                  FROM run
                  WHERE run_status = (SELECT id FROM run_status WHERE status = 'Active')"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    activeRuns = []
    for result in results:
        activeRuns.append(result[0])

    sqlQuery = """SELECT streamer.streamer_id, streamer.run_id, streamer.lumi_id,
                         streamer.events, streamer.lfn, streamer.stream_id, streamer.filesize
                  FROM streamer
                  INNER JOIN run_stream_style_assoc ON
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
                  INNER JOIN lumi_section_closed ON
                    lumi_section_closed.run_id = streamer.run_id AND
                    lumi_section_closed.lumi_id = streamer.lumi_id AND
                    lumi_section_closed.stream_id = streamer.stream_id AND
                    lumi_section_closed.close_time > 0
                  WHERE checkForZeroState(streamer.used) = 0"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    runs = set([])
    streamers = []
    for result in results:
        streamerDict = {}
        streamerDict['STREAMER_ID'] = result[0]
        streamerDict['RUN_ID'] = result[1]
        streamerDict['LUMI_ID'] = result[2]
        streamerDict['EVENTS'] = result[3]
        streamerDict['LFN'] = result[4]
        streamerDict['STREAM_ID'] = result[5]
        streamerDict['FILESIZE'] = result[6]

        if result[1] in activeRuns:
            streamerDict['CLOSEOUT'] = False
        else:
            streamerDict['CLOSEOUT'] = True

        streamers.append(streamerDict)

    return streamers

def getAvailableConversionStreamers(dbConn):
    """
    _getAvailableConversionStreamers_

    returns all unused streamers from streams
    with processing style 'Convert' or 'RegisterAndConvert'

    """
    sqlQuery = """SELECT streamer.streamer_id, streamer.run_id, streamer.lumi_id,
                         streamer.events, streamer.lfn, streamer.stream_id
                  FROM streamer
                  INNER JOIN run_stream_style_assoc ON
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id IN (
                      (SELECT id FROM processing_style WHERE name = 'Convert'),
                      (SELECT id FROM processing_style WHERE name = 'RegisterAndConvert') )
                  WHERE checkForZeroState(streamer.used) = 0"""

    dbConn.execute(sqlQuery, useTransaction = False)
    results = dbConn.fetchall()

    streamers = []
    for result in results:
        streamerDict = {}
        streamerDict['STREAMER_ID'] = result[0]
        streamerDict['RUN_ID'] = result[1]
        streamerDict['LUMI_ID'] = result[2]
        streamerDict['EVENTS'] = result[3]
        streamerDict['LFN'] = result[4]
        streamerDict['STREAM_ID'] = result[5]
        streamers.append(streamerDict)

    return streamers

def getMaxRepackLumi(dbConn, run, stream):
    """
    _getMaxRepackLumi_

    return highest lumi for which repack jobs exists
    for the given run and stream

    returns 0 if no jobs exist

    """
    sqlQuery = """SELECT MAX(a.lumi_id)
                  FROM streamer a
                  INNER JOIN repack_job_streamer_assoc b
                  ON b.streamer_id = a.streamer_id
                  WHERE a.run_id = :run
                  AND a.stream_id = ( SELECT id FROM stream WHERE name = :stream )"""

    bindVars = { "run" : run, "stream" : stream }
    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()

    if results[0][0] == None:
        return 0
    else:
        return results[0][0]

def listStreamerIDsForRepackedFiles(dbConn, fileIDs):
    """
    _listStreamerIDsForRepackedFiles_

    Given a list of repacked fileIDs, return a list of the streamer IDs 
    that they were created from.
    """
    if type(fileIDs) != list:
        fileIDs = [fileIDs]

    #if t0astFileList[0]["DATA_TIER"] != "RAW":
    #    return None
    
    sqlQuery = """SELECT streamer_id FROM streamer INNER JOIN
                  repack_streamer_assoc USING(STREAMER_ID) WHERE
                  repack_streamer_assoc.REPACKED_ID = :p_1""" 

    bindVars = []
    for fileID in fileIDs:
        bindVars.append({"p_1": fileID})

    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    streamerIDs = []    
    for result in results:
        streamerIDs.append(result[0])
        
    return streamerIDs

def countRepackStreamersWithoutChild(dbConn, run):
    """
    _countRepackStreamersWithoutChild_

    count all Bulk streamers without an EDM RAW child

    """
    sqlQuery = """SELECT COUNT(*)
                  FROM streamer
                  INNER JOIN run_stream_style_assoc ON
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
                  LEFT OUTER JOIN repack_streamer_assoc ON
                    repack_streamer_assoc.streamer_id = streamer.streamer_id
                  WHERE streamer.run_id = :RUN_ID
                  AND repack_streamer_assoc.streamer_id IS NULL
                  """

    bindVars = { "RUN_ID" : run }

    dbConn.execute(sqlQuery, bindVars, useTransaction = False)
    results = dbConn.fetchall()
    return results[0][0]

#
# method is only used by the ignoreStreamer script
#
def listSplitableStreamersByRun(dbConn, runNumber):
    """
    _listSplitableStreamersByRun_

    Retrieve a list of dictionaries containing rows from the streamer table
    that have the splitable column equal to one and belong to a specific run.
    """
    sqlQuery = """SELECT streamer.streamer_id, streamer.events, streamer.lfn
                  FROM streamer
                  INNER JOIN run_stream_style_assoc
                    run_stream_style_assoc.run_id = streamer.run_id AND
                    run_stream_style_assoc.stream_id = streamer.stream_id AND
                    run_stream_style_assoc.style_id = (SELECT id FROM processing_style WHERE name = 'Bulk')
                  WHERE checkForZeroState(streamer.used) = 0
                  AND streamer.run_id = :p_1"""

    bindVars = {"p_1": runNumber}

    dbConn.execute(sqlQuery, bindVars)
    results = dbConn.fetchall()

    streamers = []
    for result in results:
        streamerDict = {}
        streamerDict["STREAMER_ID"] = result[0]
        streamerDict["EVENTS"] = result[1]
        streamerDict["LFN"] = result[2]
        streamers.append(streamerDict)

    return streamers
