"""
_StreamerBaseHandler_

Base Job handler for taking streamers as input files.
"""

__revision__ = "$Id: StreamerBaseHandler.py,v 1.12 2010/06/22 12:00:41 hufnagel Exp $"
__version__ = "$Revision: 1.12 $"

import logging
from sets import Set

from WMCore.WMBS.File import File as WMBSFile
from WMCore.WMBS.Fileset import Fileset as WMBSFileset

from T0.Tier0Accountant.JobHandlerInterface import JobHandlerInterface

from T0.State.Database.Reader import ListJobs
from T0.State.Database.Reader import ListDatasets
from T0.State.Database.Writer import InsertJob
from T0.State.Database.Writer import InsertDataset
from T0.State.Database.Reader.ListStreamers import listStreamerIDsForRepackedFiles


class StreamerBaseHandler(JobHandlerInterface):
    """
    _StreamerBaseHandler_

    Base Job handler for taking streamers as input files.
    Currently support Repack, Express jobs 
    """
    
    def handleSuccessfulJob(self, jobInfo, jobFiles):
        """
        _handleSuccessfulJob_

        Given a list of files that were produced by a job, insert them
        into T0AST and setup all necessary associations.
        """
        jobStatus = ListJobs.listJobStatusByID(self.t0astDBConn,
                                                   jobInfo["JOB_TYPE"],
                                                   jobInfo["JOB_IDENTIFIER"])
    
        if jobStatus == None:
            msg = """ Job %s %s is not in T0AST.
                  """ % (jobInfo["JOB_TYPE"], jobInfo["JOB_IDENTIFIER"])
            logging.error(msg)
            return
        elif jobStatus == "Success":
            logging.info("Duplicate success message for %s job %s." % \
                         (jobInfo["JOB_TYPE"], jobInfo["JOB_IDENTIFIER"]))
            return
        elif jobStatus == "Abandoned":
            logging.info("Job has been abandoned: %s %s." % \
                         (jobInfo["JOB_TYPE"], jobInfo["JOB_IDENTIFIER"]))
            return
        
        #input streamer IDs for the given job not given file.
        #Since one streamer can be used to create multiple files, this will be
        #more useful values. 
        #TODO: If multiple jobs use the same streamer, it has to be handled properly.
        # i.e the same way in the wmbs jobgroup 
        inputStreamerIDs = Set()
        for t0astFile in jobFiles:
            wmbsFile = t0astFile.convertToWMBSFile()
            # insert the process dataset in the table. 
            InsertDataset.insertProcessedDataset(self.t0astDBConn, 
                                                 t0astFile["PROCESSED_DATASET"])        
            self.setAssociation(jobInfo, t0astFile)
            
            noWaitFilesetList = self.getFilesetListByFile(t0astFile)
            for fileset in noWaitFilesetList:
                fileset.addFile(wmbsFile)
                fileset.commit()
            
            if self.inputStreamerProcessingFlag:
                streamerIDs = listStreamerIDsForRepackedFiles(self.t0astDBConn, 
                                                              t0astFile["ID"])
                inputStreamerIDs = inputStreamerIDs | Set(streamerIDs)
        
        if self.inputStreamerProcessingFlag:
            self.handleInputStreamersOnSuccess(list(inputStreamerIDs))
                
        return

    def handleFailedJob(self, jobInfo):
        """
        _handleFailedJob_

        takes care of updating the job def.
        """
        jobStatus = ListJobs.listJobStatusByID(self.t0astDBConn,
                                               jobInfo["JOB_TYPE"],
                                               jobInfo["JOB_IDENTIFIER"])

        if jobStatus == None:
            logging.error("Job %s is not in T0AST." % 
                          jobInfo["JOB_IDENTIFIER"])
            return
        elif jobStatus == "Abandoned":
            logging.info("Job has been abandoned: %s." % 
                         jobInfo["JOB_IDENTIFIER"])
            return
        

    def changeState(self, jobInfo, state):
        """
        _changeState_
        
        change job state in T0AST (repack, express)
        """
        InsertJob.updateJobDef(self.t0astDBConn, jobInfo["JOB_TYPE"],
                                   jobInfo["JOB_IDENTIFIER"], state)
        return
    
    def setAssociation(self, jobInfo, t0astFile):
        """
        _setAssocitation_
        This function should set association between input files and output file
        (t0astFile). if there are other association to set (i.e. dataset or tier
        to output file, etc should implemented in child class
        """
        
        fname = "Tier0Accountant.JobHandlerInterface.StreamerBaseHandler.setAssocitation"
        
        raise NotImplementedError, fname
    
    
    def getFilesetListByFile(self, t0astFile):
        """
        _getFilesetListByFile_ 
        
        retrieve the fileset list depending on t0astFile for next processing
        The list will depends on various factor (tier, merge by size conditon)
        If the file can be added in the fileset one at a time without waiting for 
        the job group to be finished, use this list
        
        This has to be overwritten in the child class, if it needs to change default 
        behavior - returning empty list
        """
        return []
    
    def handleInputStreamersOnSuccess(self, streamerIDs):
        """
        _handleInputStreamersOnSuccess_
        
        This has to be overwritten in the child class, if something need to be done
        using input streamerIDs for the given job success.
        
        steamerIDs: set of Streamer IDs
        """
        return
    
    def inputStreamerProcessing(self):
        """
        _inputStreamerProcessing_
        
        This function just returns boolean value if there is a fileset need to be apply 
        when whole job group completed. - (If getFuntionListByStreamer is overridden 
        in sub class this should be return True).
        
        Warning: In general, this function shouldn't be overridden in the sub class.
                 However, this function won't work if the inheritance goes more than one level deep.
                 In that case, this function might be needed to be overridden
        """
        return self.__class__.__dict__.has_key("handleInputStreamersOnSuccess")

    inputStreamerProcessingFlag = property(inputStreamerProcessing)
