#!/usr/bin/env python
"""
_WMBSSBaseHandler_

Provide funcitons to common WMBS Job handling.
The sub class should be provide the appropriate fileset information
"""

__revision__ = "$Id: WMBSBaseHandler.py,v 1.41 2009/11/12 22:04:01 sryu Exp $"
__version__ = "$Revision: 1.41 $"

import logging
from sets import Set

from WMCore.WMBS.Job import Job as WMBSJob
from WMCore.WMBS.JobGroup import JobGroup as WMBSJobGroup
from WMCore.WMBS.Fileset import Fileset as WMBSFileset

from T0.State.Database.Writer import InsertDataset
from T0.Tier0Accountant.JobHandlerInterface import JobHandlerInterface

class WMBSBaseHandler(JobHandlerInterface):
    """
    _WMBSSplit_

    Job handler WMBS split jobs.
    """

    def handleSuccessfulJob(self, jobInfo, jobFiles):
        """
        _handleSuccessfulJob_

        Given a file that was returned by a prompt reco job, insert it into
        T0AST and setup all the necessary associations.        
        """
        # check jobs in WMBS
        # update WMBS job status
        # check whether job group is completed
        #     if job group is completed, release merge jobs in job queue
        #        update job group status, subscription status
        #     else do nothing
        # should be in one transaction
        
       # self.preProcessing(jobInfo, jobFiles)
        
        wmbsJob = WMBSJob(id = jobInfo["JOB_IDENTIFIER"])
        
        if not wmbsJob.exists():
            logging.error("Error Missing Job: Job with job id: %s not Exist" 
                          % jobInfo["JOB_IDENTIFIER"])
            return
        
        # only need to load to get job group id
        wmbsJob.load()
        wmbsJobGroup = WMBSJobGroup(id = wmbsJob["jobgroup"])
        # only need to load to get subscription id
        wmbsJobGroup.load()
        jobMask = wmbsJob.getMask()
        
        wmbsFileList = []
        for t0astFile in jobFiles:
            # update first/last event information for each t0ast file
            t0astFile["FIRST_EVENT"] = jobMask["FirstEvent"]
            t0astFile["LAST_EVENT"] = jobMask["LastEvent"]
            # create wmbsFile from t0ast File 
            wmbsFile = t0astFile.convertToWMBSFile()
            
            # add output will update the parentage information
            # also run lumi information
            logging.debug(" ===== parent file ids ====\n%s" % wmbsJob.getFiles(type = "id"))
            logging.debug(" ===== child file ids ====\n%s" % wmbsFile["id"])
    
            # insert into dataset table if it doesen't exist
            # if the performance is problem move this when they are created
            # i.e. runConfig and AlcaSkimInjector the first step is only needed
            # for AlcaSkim which process dataset won't be insert in AlcaSkimInjector
            InsertDataset.insertProcessedDataset(self.t0astDBConn, 
                                                 t0astFile["PROCESSED_DATASET"])
            
            InsertDataset.insertDatasetByName(self.t0astDBConn, 
                                              t0astFile["PRIMARY_DATASET"],
                                              t0astFile["PROCESSED_DATASET"],
                                              t0astFile["DATA_TIER"])

            InsertDataset.assocWMBSFileDatasetByNames(self.t0astDBConn,
                                        wmbsFile["id"],
                                        t0astFile["PRIMARY_DATASET"], 
                                        t0astFile["PROCESSED_DATASET"],
                                        t0astFile["DATA_TIER"])
            
            noWaitFilesetList = self.getFilesetListByFile(t0astFile)
            for fileset in noWaitFilesetList:
                fileset.addFile(wmbsFile)
                fileset.commit()
            wmbsFileList.append(wmbsFile)
            #add files to wmbs output
            wmbsJobGroup.addOutput(wmbsFile)
        
        # check if input files for job should be cleaned up immediately
        if self.deleteInputFiles():
            deleteFileset = WMBSFileset(name = "Deletable")
            for inputFile in wmbsJob.getFiles():
                deleteFileset.addFile(inputFile)
            deleteFileset.commit()

        self._completeJob(wmbsJob, wmbsFileList)
                
        logging.debug("WMBSBASEHandler: the wmbs job status id: Complete: %s" % wmbsJob["id"])        
        
        return wmbsJob
    
    def _completeJob(self, wmbsJob, outputFiles):
        """
        __completeJob_
        update the file parentage
        take list of wmbs file object as an argument (should contain lfn). 
        """
        if len(outputFiles) != 0:
            #only when outputFiles exist do the parentage.
            # loadData is expensive process
            wmbsJob.loadData()
            for inputFile in wmbsJob["input_files"]:
                for file in outputFiles:
                    inputFile.addChild(file["lfn"])
        
        wmbsJob["outcome"] = "success"        
        wmbsJob.save()
        wmbsJob.completeInputFiles()
        
        return
    
    def handleFailedJob(self, jobInfo):
        """
        _handleFailedJob_
        handle common task for failed wmbs job
        """
        
        wmbsJob = WMBSJob(id=jobInfo["JOB_IDENTIFIER"])
        
        if not wmbsJob.exists():
            logging.error("Job with job id: %s not Exist" % jobInfo["JOB_IDENTIFIER"])
            return

        # load the job from the database using given id (default option)
        wmbsJob.load()
        wmbsJob["outcome"] = "failure"        
        wmbsJob.save()
        wmbsJob.failInputFiles() 
        self.changeState(jobInfo, "Failure")        
        return wmbsJob
    
    
    def changeState(self, jobInfo, state):
        
        wmbsJob = WMBSJob(id = jobInfo["JOB_IDENTIFIER"])
        
        # make state change all at once it is one transaction anyway.
        # there are many transactions since there is restriction of  
        #sequence of transaction
        #TODO: we could make seperate transaction but there is no mechanism to
        #recover between the status currently.
        if state == "Success":
            self.stateChanger.propagate(wmbsJob, "executing", "created")
            self.stateChanger.propagate(wmbsJob, "complete", "executing")
            self.stateChanger.propagate(wmbsJob, "success", "complete")
            self.stateChanger.propagate(wmbsJob, "cleanout", "success")
            
        elif state == "Failure":
            self.stateChanger.propagate(wmbsJob, "executing", "created")
            self.stateChanger.propagate(wmbsJob, "jobfailed", "executing")
            self.stateChanger.propagate(wmbsJob, "exhausted", "jobfailed")
            self.stateChanger.propagate(wmbsJob, "cleanout", "exhausted")
        return
        
    def getFilesetListByFile(self, t0astFile):
        """
        _getFilesetListByFile_ 
        
        retrieve the fileset list depending on t0astFile for next processing
        The list will depends on various factor (tier, merge by size condition)
        If the file can be added in the fileset one at a time without waiting for 
        the job group to be finished, use this list
        
        This has to be overwritten in the child class, if it needs to change default 
        behavior - returning empty list
        """
        
        return []
    

    def deleteInputFiles(self):
        """
        _deleteInputFiles_ 
        
        Whether input files should be cleaned up imediately

        Default is not to cleanup, job handler can override
        """

        return False
