#!/usr/bin/env python
"""
_MergeWorkflow_

Workflow Factory for merge workflows
"""

__revision__ = "$Id: MergeWorkflow.py,v 1.36 2010/02/13 13:36:11 hufnagel Exp $"
__version__ = "$Revision: 1.36 $"

import os
import time
import logging

from ProdCommon.MCPayloads.WorkflowSpec import WorkflowSpec
import ProdCommon.MCPayloads.WorkflowTools as WorkflowTools

from T0.WorkflowFactory.FactoryInterface import FactoryInterface


class MergeWorkflow(FactoryInterface):
    """
    _MergeWorkflow_

    Util to build workflows for merge jobs
    """
    def __init__(self, runNumber, version, cmsPath, scramArch, *outModuleInfo):
        FactoryInterface.__init__(self, version, cmsPath, scramArch)
        self.run = runNumber
        self.outputModules = list(outModuleInfo)
        self.onlySites = None
        self.jobSplitParams = None
        self.workflowClass = None
        self.workflowDescription = None

    def setWorkflowSites(self, siteList):
        """
        _setWorkflowSites_
        
        Set the sites that the workflow can be run at.
        """
        self.onlySites = siteList
        return

    def setJobSplitParams(self, params):
        """
        _setJobSplitParams_
        
        Set the job splitting parameters that will be passed to the job
        splitting algorithm.
        """
        self.jobSplitParams = params
        return

    def setWorkflowClass(self, workflowClass):
        """
        _setWorkflowClass_
        
        Set the workflow class.  This is used to construct job names.
        """
        self.workflowClass = workflowClass
        return

    def setWorkflowDescription(self, description):
        """
        _setWorkflowDescription_
        
        Set the workflow description.  This is used to construct job names.
        """
        self.workflowDescription = description
        return
        
    def buildConfiguration(self, enableLazyDownload):
        """
        _buildConfiguration_

        mostly just a method to take the passed in information

        """
        outputModuleDetails = {}

        for moduleInfo in self.outputModules:

            # hack because PA handles merge jobs in a "special" way
            moduleName = "Merged"

##             moduleName = "write_%s_%s" % (moduleInfo["dataset"],
##                                           moduleInfo["dataTier"])

            outputModuleDetails[moduleName] = {
                "primaryDataset" : moduleInfo["dataset"],
                "processedDataset" : moduleInfo["processedDataset"],
                "dataTier" : moduleInfo["dataTier"],
                "acquisitionEra" : moduleInfo["acquisitionEra"],
                "processingVersion" : moduleInfo["processingVersion"],
                }

        cfgInterface = self.createConfiguration(sourceType = "PoolSource",
                                                processName = "MERGE",
                                                configName = "merge-config",
                                                enableLazyDownload = enableLazyDownload,
                                                outputModuleDetails = outputModuleDetails,
                                                noEventSort = True)

        return cfgInterface

    
    def makeWorkflowSpec(self, name, enableLazyDownload, production = True):
        """
        _makeWorkflowSpec_

        Create a workflow spec instance
        
        """
        self.workflow = WorkflowSpec()
        self.workflow.setWorkflowName(name)

        if production:
            self.workflow.setActivity("T0-Merge")
        else:
            self.workflow.setActivity("T0-Replay-Merge")
            
        self.workflow.setRequestCategory("data")
        self.workflow.setRequestTimestamp(int(time.time()))
        self.workflow.parameters["WorkflowType"] = "Merge"
        self.workflow.parameters["RequestLabel"] = name
        self.workflow.parameters["ProdRequestID"] = self.run
        self.workflow.parameters["RunNumber"] = self.run  
        self.workflow.parameters["WorkflowRunNumber"] = self.run        
        self.workflow.parameters["CMSSWVersion"] = self.cmssw["CMSSWVersion"] 
        self.workflow.parameters["ScramArch"] = self.cmssw["ScramArch"] 
        self.workflow.parameters["CMSPath"] = self.cmssw["CMSPath"]
        self.workflow.parameters["OnlySites"] = self.onlySites
        self.workflow.parameters["WorkflowClass"] = self.workflowClass
        self.workflow.parameters["WorkflowDescription"] = self.workflowDescription
                
        if self.jobSplitParams != None:
            self.workflow.parameters.update(self.jobSplitParams)

        cmsRunNode = self.workflow.payload
        cmsRunNode.name = "cmsRun1"
        cmsRunNode.type = "CMSSW"
        cmsRunNode.application["Version"] = self.cmssw["CMSSWVersion"]
        cmsRunNode.application["Executable"] = "cmsRun"
        cmsRunNode.application["Project"] = "CMSSW"
        cmsRunNode.application["Architecture"] = self.cmssw["ScramArch"]

        # runtime merge script
        cmsRunNode.scriptControls["PreExe"].append(
            "T0.Tier0Merger.RuntimeTier0Merger"
            )

        # build the configuration template for the workflow
        cmsRunNode.cfgInterface = self.buildConfiguration(enableLazyDownload)
        if cmsRunNode.cfgInterface == None:
            return None

        # generate Dataset information for workflow from cfgInterface
        for outMod,moduleInstance in cmsRunNode.cfgInterface.outputModules.items():
            primaryName = moduleInstance["primaryDataset"] 
            processedName = moduleInstance["processedDataset"] 

            outDS = cmsRunNode.addOutputDataset(primaryName,
                                                processedName,
                                                outMod)

            outDS["DataTier"] = moduleInstance["dataTier"]
            outDS["ApplicationName"] = cmsRunNode.application["Executable"]
            outDS["ApplicationFamily"] = outMod
            outDS["PhysicsGroup"] = "Tier0"
            outDS["ApplicationFamily"] = outMod

            # generate just single LFN stub (all output is merged)
            # insert them into the output module and dataset info
            outDS["LFNBase"] = self.getLFN(moduleInstance)
            moduleInstance["LFNBase"] = outDS["LFNBase"]
            moduleInstance["logicalFileName"] = os.path.join(
                outDS["LFNBase"], "%s.root" % outMod
                )

            # hack because PA handles merge jobs in a "special" way
            moduleInstance["catalog"] = "%s-Catalog.xml" % outMod

        WorkflowTools.addStageOutNode(cmsRunNode, "stageOut1")
        WorkflowTools.addLogArchNode(cmsRunNode, "logArchive")

        return self.workflow
